
var csrf = require('csurf');
var csrfProtection = csrf({ cookie: true });
var path = require('path');
var extend = require('util')._extend;
var accessTokens = require('../accessTokens');
var url = require('url');
var crypto = require('crypto');
var request = require('request');

module.exports = function (app) {

    var defaultRoutes = {
        welcome: '/welcome',
        profile: '/profile',
        login: '/login',
        sso: '/sso',
        logout: '/logout',
        register: '/register',
        activateUser: '/activateUser',
        forgotPassword: '/forgotPassword',
        resetPassword: '/resetPassword',
        unauthorized: '/unauthorized',
        confirmEmail:'/confirmEmail'
    }

    var machiModule = app.get('machiModule');
    var options = machiModule.options;
    var discourse_sso_secret = options.discourse_sso_secret;
    var discourse_api_key = options.discourse_api_key;
    var discourse_api_username = options.discourse_api_username;
    var viewsLocation = options.viewsLocation || (path.join(__dirname, '../views'));
    var routes = options.routes = extend(defaultRoutes, options.routes);

    app.get('/', function (req, res) {
        res.redirect(routes.welcome);
    });
    app.get(routes.welcome, function (req, res) {
        if (req.user)
            return res.end('Welcome ' + req.user.username);
        res.end('<a href="' + routes.login + '">Please Login </a>');
    });
    app.get(routes.login, csrfProtection, function (req, res) {

        var nextURL = req.param('next');
        if (req.user && !req.flash('status'))
            return res.redirect(nextURL || '/');
        if (req.session.returnTo == routes.logout)
            delete req.session.returnTo;
        var redirectURL = req.protocol + '://' + req.get('host') + (nextURL || req.session.returnTo || '/');
        //return res.redirect('https://machi.na:8080' + routes.login + '?next=' + redirectURL);
        renderInputs(req, res, 'login.html');
    });

    app.get(routes.sso, function (req, res) {
        if (!req.user)
            return res.redirect(routes.login + '?next=' + encodeURIComponent(req.url));
        var sso = req.param('sso');
        var sig = req.param('sig');
        var hash = crypto.createHmac('sha256', discourse_sso_secret).update(sso).digest('hex');
        if (hash != sig) return res.send('Not valid SSO token');
        sso = new Buffer(sso, 'base64').toString();
        sso = url.parse('?' + sso, true).query;
        
        var payload = {};
        payload.nonce = sso.nonce;
        payload.email = req.user.email;
        payload.external_id = req.user.username;
        payload.username = req.user.username;
        payload.name = req.user.firstName + ' ' + req.user.lastName;

        payload = url.format({ query: payload }).replace('?', '');
        payload = new Buffer(payload).toString('base64');
        var sig = crypto.createHmac('sha256', discourse_sso_secret).update(payload).digest('hex');

        var cb_url = sso.return_sso_url + '?sso=' + payload + '&sig=' + sig
        res.redirect(cb_url);
    });
 

    app.post(routes.login, csrfProtection, machiModule.middleware.auth, function (req, res) {
        var nextURL = req.param('next');
        res.redirect(nextURL || req.session.returnTo || '/');
    });




    app.get(routes.logout, machiModule.middleware.loginRequired, function (req, res) {
       
        renderInputs(req, res, 'logout.html');
    });

    app.post(routes.logout, function (req, res) {


        if (req.user) {
            var username = req.user.username;
            //Sign out from https://community.machi.na
            request.delete({
                url: 'https://community.machi.na/session/' + username, formData: {
                    api_key: discourse_api_key,
                    api_username: discourse_api_username
                }
            });
        }
        req.logout();
        var query = url.format({ query: req.query });
        res.redirect(routes.login + query);
        //res.end('Logged out');
    });


    var profileFields =
        [{
            name: 'username',
            label: 'Username',
            required: true,
            readonly: true,
            placeholder: 'Username',
            type: 'text'
        }, {
            name: 'firstName',
            label: 'First Name*',
            required: true,
            placeholder: 'First Name',
            type: 'text'
        }, {
            name: 'lastName',
            label: 'Last Name*',
            required: true,
            placeholder: 'Last Name',
            type: 'text'
        }, {
            name: 'email',
            label: 'Email*',
            required: true,
            placeholder: 'Email',
            type: 'text'
        }, {
            name: 'mobilePhone',
            label: 'Phone',
            placeholder: 'Mobile Phone',
            type: 'text'
        }, {
            name: 'title',
            label: 'Title',
            placeholder: 'Title',
            type: 'text'
        }, {
            name: 'organization',
            label: 'Organization',
            placeholder: 'Organization',
            type: 'text'
        }, {
            name: 'bestTimeToCall',
            label: 'Best time to call',
            customControl: function (selectedValue, readonly) {
                if (readonly) return '<input type="text" class="form-control" readonly value="' + selectedValue+'"/>';
                var options = ['', 'Daytime', 'Evening'];
                var html = options.map(o => {
                    var t = '<option value="' + o + '" ' + (o == selectedValue ? 'selected' : '') + '>' + o + '</option>';
                    return t;
                }).join('\n');
                return '<select class="form-control" name="bestTimeToCall" ' + (readonly ? 'readonly disabled' : '') + '>' + html + '</select>';
            }
        }, {
            name: 'publicProfile',
            label: 'Share my profile and contact information with sponsors.',
            type: 'checkbox',
            showInPublicProfile: false
        }];
    var registerFields =
        [{
            name: 'firstName',
            label: 'First Name*',
            required: true,
            placeholder: 'First Name',
            type: 'text'
        }, {
            name: 'lastName',
            label: 'Last Name*',
            required: true,
            placeholder: 'Last Name',
            type: 'text'
        }, {
            name: 'email',
            label: 'Email*',
            required: true,
            placeholder: 'Email',
            type: 'text'
        }, {
            name: 'password',
            label: 'Password*',
            required: true,
            placeholder: 'Password',
            type: 'password'
        }];
    //For public profile only
    app.get(routes.profile + '/:username', csrfProtection, function (req, res, next) {
        var username = req.param('username');
        if (!username) return next();
        machiModule.connect.getUser(username, function (err, account) {
            if (err) return next();
            var profile = account.profile;
            if (!profile.publicProfile && (!req.user || profile.username != req.user.username)) return next();
            renderInputs(req, res, 'profile.html', { fields: profileFields, profile: profile, readonly: true });
        });

    });

    app.get(routes.profile, csrfProtection, machiModule.middleware.loginRequired, function (req, res) {
        machiModule.connect.getUser(req.user.username, function (err, account) {
            if (err) {
                console.error(err);
                return res.end('Failed to load user\'s profile');
            }
            var profile =req.flash('form')|| account.profile;
            renderInputs(req, res, 'profile.html', { fields: profileFields, profile: profile });
        });
        
    });

    app.post(routes.profile, function (req, res) {
        var form = req.body;
        if (typeof (form) != 'object')
            return res.end('Invalid request');
        var formKeys = Object.keys(form);
        var newEmail = form.email !== req.user.email;
        //Sets the checkbox values with false in case it's not checked
        profileFields.filter(f => f.type == 'checkbox').forEach(f => {
            form[f.name] = formKeys.indexOf(f.name) >= 0;
        });
        machiModule.utilities.updateProfile(req, form, function (err, newProfile) {
            var feedbacks = [];
            if (!err) {
                feedbacks.push('Your profile has been updated');
                if (newEmail) feedbacks.push('Please check your ' + form.email + ' inbox and confirm it');
                form = undefined;
            }
            redirectWithFlash(req, res, routes.profile, err, { form: form, feedbacks: feedbacks });
        });

    });


    app.get(routes.confirmEmail, function (req, res) {
        var token = req.param('token');
        if (!token) {
            return req.end('Token not valid');
        }
        machiModule.tokens.getToken('ConfirmEmail', token, false, function (err, data) {
            if (err) return res.end('Token not valid');
            machiModule.connect.getUser(data.username, function (err, user) {
                if (err) return res.end(err);
                var profile = user.profile;
                profile.email = data.newEmail;
                machiModule.connect.updateUser(profile.username, profile, function (err, newUser) {
                    req.session.passport.user = newUser.profile;
                    redirectWithFlash(req, res, routes.profile, err, { form: profile, feedbacks: ['Your email address has been verified'] });
                });
            });
            
        });
    });

    app.get(routes.register, function (req, res) {
        renderInputs(req, res, 'register.html', { fields: registerFields });
    });


    app.post(routes.register, function (req, res) {
        var form = req.body;
        machiModule.connect.addUser({ firstName: form.firstName, lastName: form.lastName, email: form.email }, form.password, function (err, user) {

            if (err) {
                return redirectWithFlash(req, res, routes.register, err, { form: form });
            }
            //User created, go to verfiy page
            renderInputs(req, res, 'verify-email.html', { form: form });
        });

    });


    app.get(routes.activateUser, function (req, res) {
        var token = req.param('token') || '';
        machiModule.tokens.getToken('User', token, true, function (err, emailAddress) {
            if (err) return res.end('Token not valid');
            machiModule.connect.activateUser(emailAddress, true, function (err) {
                if (err) return res.end(err);
                //User has been activated, go to login page
                req.session.returnTo = '/';
                redirectWithFlash(req, res, routes.login, undefined, { status: 'verified', form: { email: emailAddress } });
            });
        });
    });

    app.get(routes.unauthorized, function (req, res) {
        renderInputs(req, res, 'unauthorized.html');
    });
    app.get(routes.forgotPassword, csrfProtection, function (req, res) {
        renderInputs(req, res, 'forgot-password.html', { email: req.param('email') });
    });

    app.post(routes.forgotPassword, csrfProtection, function (req, res) {
        var form = req.body;
        machiModule.connect.getUserByEmail(form.email, function (err, user) {
            if (err)
                return redirectWithFlash(req, res, routes.forgotPassword, err);
            var profile = user.profile;
            if (user.status == 'STAGED') {
                machiModule.connect.sendActivationEmail(profile.username, function (err) {
                    renderInputs(req, res, 'verify-email.html', { form: form });
                });
            } else {
                machiModule.connect.sendForgotPasswordEmail(profile.username, function (err) {
                    if (err) return res.end(err);
                    //User has been activated, go to login page
                    redirectWithFlash(req, res, routes.login, undefined, { status: 'forgot' });
                });
            }
        });

    });

    app.get(routes.resetPassword, csrfProtection, function (req, res) {
        var token = req.param('token') || '';
        machiModule.tokens.getToken('Password', token, false, function (err, emailAddress) {
            if (err) return res.end('Token not valid');
            renderInputs(req, res, 'change-password.html', { resetToken: token });
        });
    });

    app.post(routes.resetPassword, csrfProtection, function (req, res) {
        var resetToken = req.body.resetToken;
        var passwordNew = req.body.passwordNew;
        var passwordAgain = req.body.passwordAgain;
        if (!resetToken || !passwordNew || passwordNew != passwordAgain) {
            return redirectWithFlash(req, res, routes.resetPassword + '?token=' + resetToken, 'Password does not match');
        }
        machiModule.tokens.getToken('Password', resetToken, false, function (err, emailAddress) {
            if (err) return res.end('Token not valid');
            //Reset the user password
            machiModule.connect.changeUserPassword(emailAddress, passwordNew, true, function (err) {
                if (err) {
                    return redirectWithFlash(req, res, routes.resetPassword + '?token=' + resetToken, err);
                }
                machiModule.tokens.removeToken('Password', resetToken);
                //User has been activated, go to login page
                redirectWithFlash(req, res, routes.login, undefined, { status: 'reset' });
            });
        });
    });


    app.all('/me', machiModule.middleware.apiAuthenticationRequired, function (req, res) {
        res.json({ account: req.user });
    });

    app.all('/tokens/requestToken', machiModule.middleware.apiAuthenticationRequired, function (req, res) {
        var token = accessTokens.createToken(req, {}, 60*60*24*7);
        res.json(token);
    });

    function redirectWithFlash(req, res, url, err, flashObj) {

        if (err && Array.isArray(err)) req.flash('errors', err);
        if (err && !Array.isArray(err)) req.flash('error', err);

        if (typeof (flashObj) == 'object') {
            Object.keys(flashObj).forEach(function (k) {
                req.flash(k, flashObj[k]);
            });
        }
        res.redirect(url);

    }

    function renderInputs(req, res, view, locals) {
        locals = locals || {};
        locals.csrfToken = req.csrfToken && req.csrfToken();

        var flashs = (req.flash && req.flash()) || [];
        Object.keys(flashs).forEach(function (key) {
            locals[key] = flashs[key];
        });
        if (!locals.form)
            locals.form = locals.body;
        if (req.user)
            locals.user = req.user;

        locals.routes = routes;
        var filename = path.join(viewsLocation, view);
        res.render(filename, {
            inputs: locals
        });
    }

}