﻿var OktaAPI = require('oktajs');
var MachiConnect  = function (app) {
    this.init(app);
}

MachiConnect.prototype = {
    /**
     * Initlize the Connector instance
     *
     * @method
     *
     * @param {Express} app - The express app
     */
    init: function (app) {
        var instance = this;
        this.app = app;
        var machiModule=instance.machiModule= app.get('machiModule');
       var options = machiModule.options;

       var okta = new OktaAPI(options.apiKey, options.oktaDomain, options.oktapreview);
        instance.oktaUsers = okta.users;
        instance.oktaGroups = okta.groups;
        instance.oktaSessions = okta.sessions;
        instance.oktaApps = okta.apps;

        instance.emails = machiModule.emails;
        instance.options = options;

        //The token link expire in 24 hours
        instance.forgotPasswordToken_Expire = options.forgotPasswordToken_Expire || 60 * 60 * 24;
        //The token link expire in 7 days
        instance.activationToken_Expire = options.activationToken_Expire || 60 * 60 * 24 * 7;
    },
    /**
     * Make sure the profile has only the valid properties for register and update api calls
     */
    makeProfile: function (profile) {

        var newProfile = JSON.parse(JSON.stringify(profile));

        // Create a profile object that Okta can recognize
        if (newProfile.username && newProfile.username.indexOf('@') < 0) {
            var emailEnd = newProfile.email && newProfile.email.split('@')[1];
            if (!emailEnd) emailEnd = 'username.com'
            newProfile.username = newProfile.username + '@' + emailEnd;
        }
        newProfile.login = newProfile.username || newProfile.email;

        delete newProfile.id;
        delete newProfile.username;
        return newProfile;
    },
    /**
     * Creates a new user that reflects the structure of an Okta User Object
     * @method addUser
     * @param firstName
     * @param lastName
     * @param email
     * @param username optional
     * @param mobilePhone optional
     * @param customAttribs a map of property -> propertyValue for custom attributes on a user
     * @returns {{}}
     */
    addUser: function (profile, password, callback) {
        var instance = this;
        profile = instance.makeProfile(profile);
        // Create a credentials object 
        var newCreds = instance.oktaUsers.helpers.constructCredentials(password);

        var username = profile.login.split('@')[0];
        instance.getUser(username, function (err, user) {
            if (!err && user)
                return callback('Username ' + username + ' already exists. please use differnt email address');
            // Send the request off to Okta 
            instance.oktaUsers.add(profile, newCreds, false, function (data) {
                oktaReply(data, function (err, user) {
                    if (err) return callback(err);
                    instance.sendActivationEmail(user.profile.login, function (err) {
                        callback(err, user);
                    });
                });
            });
        });
        
    },
    /**
     * Sends an email that has a link to activate this User
     * @method sendActivationEmail
     * @param username
     * @returns {{}}
     */
    sendActivationEmail: function (username, callback) {
        var instance = this;
        var tokens = instance.machiModule.tokens;
        var activationUrl = instance.options.baseURL + instance.options.routes.activateUser + '?token=' + tokens.createToken(TokenKeys.User, username, instance.activationToken_Expire);
        instance.emails.sendEmail(username, 'verify', {
            activationUrl: activationUrl
        }, callback);
    },

    /**
     * Updat user's profile
     * @method updateUser
     * @param username
     * @param profile
     * @returns {{}}
     */
    updateUser: function (username, profile, callback) {
        var instance = this;
        profile.username = username;
        profile = instance.makeProfile(profile);
        // Send the request off to Okta 
        instance.oktaUsers.update(username, profile, undefined, function (data) {
            oktaReply(data, callback);
        });
    },
    /**
     * Updat user's profile and password
     * @method changeUserPassword
     * @param username
     * @param password
     * @param sendEmail
     */
    changeUserPassword: function (username, password, sendEmail, callback) {
        var instance = this;
        var newCreds = instance.oktaUsers.helpers.constructCredentials(password);
        // Send the request off to Okta 
        instance.oktaUsers.update(username, undefined, newCreds, function (data) {
            oktaReply(data, function (err) {
                if (err || !sendEmail) return callback(err);
                instance.emails.sendEmail(username, 'passwordChanged', {}, callback);
            });
        });
    },

    /**
     * Serach for a user
     * @method getUser
     * @param username: The user id
     * @param callback
     * @returns {{}}
     */
    getUser: function (username, callback) {
        var instance = this;
        // Send the request off to Okta 
        instance.oktaUsers.get(username, function (data) {
            oktaReply(data, callback);
        });

    },

    /**
     * Serach for a user by email address
     * @method getUserByEmail
     * @param email: The user email address
     * @param callback
     * @returns {{}}
     */
    getUserByEmail: function (email, callback) {
        var instance = this;
        email = (email || '').toLowerCase();
        // Send the request off to Okta 
        instance.oktaUsers.list({ q: email }, false, function (data) {
            oktaReply(data, function (err, arr) {
                if (err) return callback(err);
                arr = (arr || []).filter(c => c.profile.email.toLowerCase() == email);
                if (!arr || !arr.length)
                    return callback('User with email address "' + email + '" does not exists');
                oktaReply({ resp: arr[0] }, callback);
            });
        });

    },
    /**
     * Serach for a user by filter text, see https://developer.okta.com/docs/api/resources/users.html#list-users-with-a-filter
     * @method getUsers
     * @param queryObj: The query obj, can have search,filter,q
     * @param callback
     * @returns {{}}
     */
    getUsers: function (queryObj, callback) {
        var instance = this;
        // Send the request to Okta 
        instance.oktaUsers.list(queryObj, true, function (data) {
            oktaReply(data, callback);
        });
    },
    /**
     * Activate a new user 
     * @method activateUser
     * @param username: The user id
     * @param sendEmail: should send an email
     * @param callback
     * @returns {{}}
     */
    activateUser: function (username, sendEmail, callback) {
        var instance = this;
        // Send the request off to Okta 
        instance.oktaUsers.activate(username, false, function (data) {
            oktaReply(data, function (err) {
                if (err) return callback(err);
                if (sendEmail)
                    instance.emails.sendEmail(username, 'welcome', {}, callback);
                else callback();
            });
        });

    },


    /**
     * Reset User's password link
     * @method sendForgotPasswordEmail
     * @param username: The user id
     * @param callback
     * @returns {{}}
     */
    sendForgotPasswordEmail: function (username, callback) {
        var instance = this;
        var tokens = instance.machiModule.tokens;
        var url = instance.options.baseURL + instance.options.routes.resetPassword + '?token=' + tokens.createToken(TokenKeys.Password, username, instance.forgotPasswordToken_Expire);
        instance.emails.sendEmail(username, 'resetPassword', {
            resetPasswordURL: url
        }, function (err) {
            callback(err);
        });

    },
    
    /**
     * Gets all groups assigned to this user
     * @method getUserGroups
     * @param username: The user id
     * @param callback
     * @returns {{}}
     */
    getUserGroups: function (username, callback) {
        var instance = this;

        // Send the request off to Okta 
        instance.oktaUsers.getGroups(username, function (data) {
            oktaReply(data, callback);
        });

    },
    /**
     * check if this user/pass are valid, if so, create a session 
     *
     * @method
     *
     * @param {function} callback  - The function that will hold the two parameters
     *                               err: the error object in case something error
     *                               session: the session object in case no errors
     */
    createSession: function (username, password, callback) {
        var instance = this;

        instance.oktaSessions.create(username, password, {}, function (data) {
            oktaReply(data, function (err, session) {
                if (err) return callback(err);
                instance.getUser(session.username, function (err, user) {
                    session.user = user;
                    callback(err, session);
                });
            });
        });

    },
    /**
     * check if this sessionId is valid
     *
     * @method
     *
     * @param {function} callback  - The function that will hold the two parameters
     *                               err: the error object in case something error
     *                               session: the session object in case no errors
     */
    validateSession: function (sessionId, callback) {
        var instance = this;

        if (!sessionId) return callback('SessionId cannot be null');
        instance.oktaSessions.validate(sessionId, function (data) {
            oktaReply(data, function (err, session) {
                if (err) return callback(err);
                instance.getUser(session.username, function (err, user) {
                    session.user = user;
                    callback(err, session);
                });
            });
        });

    },
    /**
     * Extend this sessionId
     *
     * @method
     *
     * @param {function} callback  - The function that will hold the two parameters
     *                               err: the error object in case something error
     *                               session: the session object in case no errors
     */
    extendSession: function (sessionId, callback) {
        var instance = this;

        instance.oktaSessions.extend(sessionId, function (data) {
            oktaReply(data, callback);
        });

    },
    /**
     * CLose this sessionId
     *
     * @method
     *
     * @param {function} callback  - The function that will hold the two parameters
     *                               err: the error object in case something error
     *                               session: the session object in case no errors
     */
    closeSession: function (sessionId, callback) {
        var instance = this;

        instance.oktaSessions.close(sessionId, function (data) {
            oktaReply(data, callback);
        });

    },

    /**
     * List all groups
     *
     * @method
     *
     * @param {String} search - The group name or search query
     * @param {function} callback  - The function that will hold the two parameters
     *                               err: the error object in case something error
     *                               group: the group object in case no errors
     */
    listGroups: function (search, callback) {
        var instance = this;

        instance.oktaGroups.list(search, true, function (data) {
            oktaReply(data, callback);
        });
    },
    /**
     * get group's users
     *
     * @method
     *
     * @param {String} groupName - The group name
     * @param {function} callback  - The function that will hold the two parameters
     *                               err: the error object in case something error
     *                               group: the group object in case no errors
     */
    getGroupUsers: function (groupName, callback) {
        var instance = this;

        instance.listGroups(groupName, function (err, groups) {
            if (err) return callback(err);
            if (!groups[0]) return callback('Group not found');

            instance.oktaGroups.getUsers(groups[0].id, {}, true, function (data) {
                oktaReply(data, callback);
            });

        });

    },
    /**
     * Add a user to this group
     *
     * @method
     *
     * @param {String} groupName - The group name
     * @param {function} callback  - The function that will hold the parameters
     *                                   err: the error object in case something error
     */
    AddGroupUser: function (groupName, username, callback) {
        var instance = this;

        instance.listGroups(groupName, function (err, groups) {
            if (err) return callback(err);
            if (!groups[0]) return callback('Group not found');

            instance.getUser(username, function (err, user) {
                if (err) return callback(err);
                instance.oktaGroups.addUser(groups[0].id, user.id, function (data) {
                    oktaReply(data, callback);
                });

            });

        });
    },
    //reply with bad auth header user/pass
    logout: function (req, res, redirectToIndex) {
        var instance = this;

        if (req.session && req.session.destroy)
            req.session.destroy();
        if (req.cookies) {
            Object.keys(req.cookies).forEach(function (name) {
                res.clearCookie(name);
            });
        }
        if (redirectToIndex)
            res.redirect('/');
    }



}


function oktaReply(oktaResponse, callback) {
    if (!oktaResponse) return callback();
    if (!oktaResponse.error) {
        var r = oktaResponse.resp;
        if (r.profile && r.profile.login) {
            r.profile.username = r.profile.login.split('@')[0];
            r.profile.id = r.id;
        }
        if (r.login)
            r.username = r.login.split('@')[0];
        return callback(undefined, r);
    }

    var data = oktaResponse.resp;
    if (!data) return callback(oktaResponse.error);

    try {
        data = JSON.parse(data);
        var errorCauses = data.errorCauses || [];
        var errors = errorCauses.map(function (e) {
            return e.errorSummary
        });
        if (errorCauses.length)
            callback(errors);
        else callback(data.errorSummary);
    } catch (ex) {
        callback(oktaResponse.error);
    }
}


module.exports.MachiConnect = MachiConnect;