
var csrf = require('csurf');
var csrfProtection = csrf({ cookie: true });
var path = require('path');
var extend = require('util')._extend;
var accessTokens = require('../accessTokens');
var url = require('url');
var crypto = require('crypto');
var request = require('request');

module.exports = function (app) {

    var defaultRoutes = {
        welcome: '/welcome',
        profile: '/profile',
        login: '/login',
        sso: '/sso',
        logout: '/logout',
        register: '/register',
        activateUser: '/activateUser',
        forgotPassword: '/forgotPassword',
        resetPassword: '/resetPassword',
        unauthorized: '/unauthorized',
        confirmEmail: '/confirmEmail',
        twoFactor: '/2factor',
        requestToken:'/tokens/requestToken'
    }

    var machiModule = app.get('machiModule');
    var options = machiModule.options;
    var discourse_sso_secret = options.discourse_sso_secret;
    var discourse_api_key = options.discourse_api_key;
    var discourse_api_username = options.discourse_api_username;
    var viewsLocation = options.viewsLocation || (path.join(__dirname, '../views'));
    var routes = options.routes = extend(defaultRoutes, options.routes);

    app.get('/', function (req, res) {
        res.redirect(routes.welcome);
    });
    app.get(routes.welcome, function (req, res) {
        if (req.user)
            return res.end('Welcome ' + req.user.username);
        res.end('<a href="' + routes.login + '">Please Login </a>');
    });
    app.get(routes.login, csrfProtection, function (req, res) {

        var nextURL = req.param('next');
        if (req.user && !req.flash('status'))
            return afterLogin(req, res);
        if (req.session.returnTo == routes.logout)
            delete req.session.returnTo;
        renderInputs(req, res, 'login.html');
    });

    app.get(routes.sso, function (req, res) {
        if (!req.user)
            return res.redirect(routes.login + '?next=' + encodeURIComponent(req.url));
        var sso = req.param('sso');
        var sig = req.param('sig');
        var hash = crypto.createHmac('sha256', discourse_sso_secret).update(sso).digest('hex');
        if (hash != sig) return res.send('Not valid SSO token');
        sso = new Buffer(sso, 'base64').toString();
        sso = url.parse('?' + sso, true).query;
        
        var payload = {};
        payload.nonce = sso.nonce;
        payload.email = req.user.email;
        payload.external_id = req.user.username;
        payload.username = req.user.username;
        payload.name = req.user.display_name;

        payload = url.format({ query: payload }).replace('?', '');
        payload = new Buffer(payload).toString('base64');
        var sig = crypto.createHmac('sha256', discourse_sso_secret).update(payload).digest('hex');

        var cb_url = sso.return_sso_url + '?sso=' + payload + '&sig=' + sig
        res.redirect(cb_url);
    });
 
    function authorizeLogin(req, res, next) {
        machiModule.connect.validateLogin(req.param('username'), req.param('password'), (validateError, user) => {
            if (validateError && user && user.status == 'password reset') {
                machiModule.connect.createForgotPasswordLink(user.username, (err, url) => {
                    if (err) return next(err);
                    redirectWithFlash(req, res, url, validateError);
                });
                return;
            }
            if (validateError && user && user.status == 'staged') {
                machiModule.connect.sendActivationEmail(user.username, (err) => {
                    if (err) return next(err);
                    redirectWithFlash(req, res, routes.login, validateError);
                });
                return
            }
            if (!validateError) {
                if (user.two_factor_enabled) {
                    if (!machiModule.twoFactorAuth.verify(user.two_factor_secret, req.param('code')))
                        return redirectWithFlash(req, res, req.url, '2Factor authentication required', { form: req.body, twoFactorRequired: true });
                }
            }
            machiModule.middleware.auth(req, res, next);
        });
    }
    function afterLogin(req, res) {
        var nextURL = req.param('next') || req.session.returnTo || '/';
        var sendToken = req.param('sendToken');
        //if sendToken then send the accessToken to that url
        if (sendToken) {
            var token = accessTokens.createJwtToken(req, {}, 60 * 60 * 24 * 7);
            var data = {
                token: token,
                user: options.baseURL + '/me?accessToken=' + token.accessToken,
                refreshToken: options.baseURL + routes.requestToken + '?accessToken=' + token.accessToken
            };
            data = encodeURIComponent(JSON.stringify(data));
            if (nextURL.indexOf('?') < 0)
                nextURL += '?loginData=' + data;
            else nextURL += '&loginData=' + data;
        }
        res.redirect(nextURL);
    }
    app.post(routes.login, csrfProtection, authorizeLogin, afterLogin);

    app.get(routes.twoFactor, machiModule.middleware.loginRequired, function (req, res,next) {

        if (req.user.two_factor_enabled) {

            renderInputs(req, res, 'two-factor.html', { enabled: true });

        } else {

            var secret = req.flash('secret') || machiModule.twoFactorAuth.generateSecret();
            machiModule.twoFactorAuth.generateQRImage(secret, (err, png) => {
                if (err) return next(err);
                renderInputs(req, res, 'two-factor.html', { secret: secret, png: png });
            });
        }

    });
    app.post(routes.twoFactor, function (req, res, next) {
        if (req.param('fun') == 'enable') {
            var secret = req.param('secret');
            var code = req.param('code');
            //if it's already set then don't do anything
            if (req.user.two_factor_enabled)
                return redirectWithFlash(req, res, routes.twoFactor, '2Factor Authentication already enabled', { secret: secret });

            var verified = secret && code && machiModule.twoFactorAuth.verify(secret, code);
            if (!verified)
                return redirectWithFlash(req, res, routes.twoFactor, 'Enabling Authentication failed', { secret: secret });

            req.user.two_factor_secret = secret;
            req.user.two_factor_enabled = true;
            machiModule.utilities.updateProfile(req, req.user, function (err) {
                if (err) return redirectWithFlash(req, res, routes.twoFactor, err, { secret: secret });
                redirectWithFlash(req, res, routes.twoFactor, undefined, { feedbacks: ['2Factor Authentication has been enabled'] });
            });
        } else if (req.param('fun') == 'disable') {
            var code = req.param('code');
            if (!req.user.two_factor_enabled)
                return redirectWithFlash(req, res, routes.twoFactor, '2Factor Authentication already disabled', {});

            var verified = code && machiModule.twoFactorAuth.verify(req.user.two_factor_secret, code);
            if (!verified)
                return redirectWithFlash(req, res, routes.twoFactor, 'Disabling Authentication failed', {});

            req.user.two_factor_enabled = false;
            machiModule.utilities.updateProfile(req, req.user, function (err) {
                if (err) return redirectWithFlash(req, res, routes.twoFactor, err, { secret: secret });
                redirectWithFlash(req, res, routes.twoFactor, undefined, { feedbacks: ['2Factor Authentication has been disabled'] });
            });
        }
        else next('Unknown')

    });

    app.get(routes.logout, machiModule.middleware.loginRequired, function (req, res) {
       
        renderInputs(req, res, 'logout.html');
    });

    app.post(routes.logout, function (req, res) {


        if (req.user) {
            var username = req.user.username;
            //Sign out from https://community.machi.na
            request.delete({
                url: 'https://community.machi.na/session/' + username, formData: {
                    api_key: discourse_api_key,
                    api_username: discourse_api_username
                }
            });
        }
        req.logout();
        var query = url.format({ query: req.query });
       // res.redirect(routes.login + query);
        res.redirect('/');
        //res.end('Logged out');
    });


    var profileFields =
        [{
            name: 'username',
            label: 'Username',
            required: true,
            readonly: true,
            placeholder: 'Username',
            type: 'text'
        }, {
            name: 'display_name',
            label: 'Display Name*',
            required: true,
            placeholder: 'Display Name',
            type: 'text'
        }, {
            name: 'email',
            label: 'Email*',
            required: true,
            placeholder: 'Email',
            type: 'text'
        }, {
            name: 'phone_number',
            label: 'Phone',
            placeholder: 'Mobile Phone',
            type: 'text'
        }, {
            name: 'title',
            label: 'Title',
            placeholder: 'Title',
            type: 'text'
        }, {
            name: 'organization',
            label: 'Organization',
            placeholder: 'Organization',
            type: 'text'
        }, {
            name: 'best_time_to_call',
            label: 'Best time to call',
            customControl: function (selectedValue, readonly) {
                if (readonly) return '<input type="text" class="form-control" readonly value="' + selectedValue + '"/>';
                var options = ['', 'Daytime', 'Evening'];
                var html = options.map(o => {
                    var t = '<option value="' + o + '" ' + (o == selectedValue ? 'selected' : '') + '>' + o + '</option>';
                    return t;
                }).join('\n');
                return '<select class="form-control" name="best_time_to_call" ' + (readonly ? 'readonly disabled' : '') + '>' + html + '</select>';
            }
        }, {
            label: '2Factor',
            showInPublicProfile: false,
            customControl: function (selectedValue, readonly, inputs) {
                return '<a class="btn btn-default" href="/2factor">Manage 2Factor Authorization</a>';
            }
        }, {
            name: 'is_public',
            label: '',
            showInPublicProfile: false,
            customControl: function (selectedValue, readonly, inputs) {
                return '<div class="checkbox"><label><input name="is_public" type="checkbox" ' + (selectedValue ? 'checked' : '') + ' ' + (readonly ? 'readonly disabled' : '') + '> Share my profile and contact information with sponsors.</label></div>';
            }
        }];
    var registerFields =
        [{
            name: 'display_name',
            label: 'Display Name*',
            required: true,
            placeholder: 'Display Name',
            type: 'text'
        }, {
            name: 'username',
            label: 'Username*',
            required: true,
            placeholder: 'Username',
            type: 'text'
        }, {
            name: 'email',
            label: 'Email*',
            required: true,
            placeholder: 'Email',
            type: 'text'
        }, {
            name: 'password',
            label: 'Password*',
            required: true,
            placeholder: 'Password',
            type: 'password'
        }];
    //For public profile only
    app.get(routes.profile + '/:username', csrfProtection, function(req, res, next) {
        var username = req.param('username');
        if (!username) return next();
        machiModule.connect.getUser(username, function(err, user) {
            if (err) return next();
            if (!user.is_public && (!req.user || user.username != req.user.username)) return next();
            renderInputs(req, res, 'profile.html', { fields: profileFields, profile: user, readonly: true });
        });

    });

    app.get(routes.profile, csrfProtection, machiModule.middleware.loginRequired, function (req, res) {
        machiModule.connect.getUser(req.user.username, function (err, user) {
            if (err) {
                console.error(err);
                return res.end('Failed to load user\'s profile');
            }
            renderInputs(req, res, 'profile.html', { fields: profileFields, profile: (req.flash('form')|| user) });
        });
        
    });

    app.post(routes.profile, function (req, res) {
        var form = req.body;
        if (typeof (form) != 'object')
            return res.end('Invalid request');
        var formKeys = Object.keys(form);
        var newEmail = form.email !== req.user.email;
        //Sets the checkbox values with false in case it's not checked
        profileFields.filter(f => f.type == 'checkbox').forEach(f => {
            form[f.name] = formKeys.indexOf(f.name) >= 0;
        });
        machiModule.utilities.updateProfile(req, form, function (err, newProfile) {
            var feedbacks = [];
            if (!err) {
                feedbacks.push('Your profile has been updated');
                if (newEmail) feedbacks.push('Please check your ' + form.email + ' inbox and confirm it');
                form = undefined;
            }
            redirectWithFlash(req, res, req.url, err, { form: form, feedbacks: feedbacks });
        });

    });


    app.get(routes.confirmEmail, function (req, res) {
        var token = req.param('token');
        if (!token) {
            return req.end('Token not valid');
        }
        machiModule.tokens.getToken(TokenKeys.ConfirmEmail, token, false, function (err, data) {
            if (err) return res.end('Token not valid');
            machiModule.connect.getUser(data.username, function (err, user) {
                if (err) return res.end(err);
                user.email = data.newEmail;
                user.email_confirmed = true;
                machiModule.connect.updateUser(user.username, user, function (err, newUser) {
                    req.session.passport.user = newUser;
                    redirectWithFlash(req, res, routes.profile, err, { form: user, feedbacks: ['Your email address has been verified'] });
                });
            });
            
        });
    });

    app.get(routes.register, function (req, res) {
        renderInputs(req, res, 'register.html', { fields: registerFields });
    });


    app.post(routes.register, function (req, res) {
        var form = req.body;
        machiModule.connect.addUser(form, function(err, user) {

            if (err) {
                return redirectWithFlash(req, res, routes.register, err, { form: form });
            }
            //User created, go to verfiy page
            renderInputs(req, res, 'verify-email.html', { form: form });
        });

    });


    app.get(routes.activateUser, function (req, res) {
        var token = req.param('token') || '';
        machiModule.tokens.getToken(TokenKeys.User, token, true, function (err, emailAddress) {
            if (err) return res.end('Token not valid');
            machiModule.connect.activateUser(emailAddress, true, function (err) {
                if (err) return res.end(err);
                //User has been activated, go to login page
                req.session.returnTo = '/';
                redirectWithFlash(req, res, routes.login, undefined, { status: 'verified', form: { email: emailAddress } });
            });
        });
    });

    app.get(routes.unauthorized, function (req, res) {
        renderInputs(req, res, 'unauthorized.html');
    });
    app.get(routes.forgotPassword, csrfProtection, function (req, res) {
        renderInputs(req, res, 'forgot-password.html', { email: req.param('email') });
    });

    app.post(routes.forgotPassword, csrfProtection, function (req, res) {
        var form = req.body;
        machiModule.connect.getUser(form.email, function (err, user) {
            if (err)
                return redirectWithFlash(req, res, routes.forgotPassword, err);
            if (!user.email_confirmed) {
                machiModule.connect.sendActivationEmail(user.username, function (err) {
                    renderInputs(req, res, 'verify-email.html', { form: form });
                });
            } else {
                machiModule.connect.sendForgotPasswordEmail(user.username, function (err) {
                    if (err) return res.end(err);
                    //User has been activated, go to login page
                    redirectWithFlash(req, res, routes.login, undefined, { status: 'forgot' });
                });
            }
        });

    });

    app.get(routes.resetPassword, csrfProtection, function (req, res) {
        var token = req.param('token') || '';
        machiModule.tokens.getToken(TokenKeys.Password, token, false, function (err, emailAddress) {
            if (err) return res.end('Token not valid');
            renderInputs(req, res, 'change-password.html', { resetToken: token });
        });
    });

    app.post(routes.resetPassword, csrfProtection, function (req, res) {
        var resetToken = req.body.resetToken;
        var passwordNew = req.body.passwordNew;
        var passwordAgain = req.body.passwordAgain;
        if (!resetToken || !passwordNew || passwordNew != passwordAgain) {
            return redirectWithFlash(req, res, routes.resetPassword + '?token=' + resetToken, 'Password does not match');
        }
        machiModule.tokens.getToken(TokenKeys.Password, resetToken, false, function (err, emailAddress) {
            if (err) return res.end('Token not valid');
            //Reset the user password
            machiModule.connect.changeUserPassword(emailAddress, passwordNew, true, function (err) {
                if (err) {
                    return redirectWithFlash(req, res, routes.resetPassword + '?token=' + resetToken, err);
                }
                machiModule.tokens.removeToken(TokenKeys.Password, resetToken, err => {
                    //User has been activated, go to login page
                    redirectWithFlash(req, res, routes.login, undefined, { status: 'reset' });
                });
                
            });
        });
    });


    app.all('/me', machiModule.middleware.apiAuthenticationRequired, function (req, res) {
        var user = JSON.parse(JSON.stringify(req.user));
        delete user.two_factor_secret;
        delete user.password_hash;
        
        res.json({ account: user });
    });

    app.all(routes.requestToken, machiModule.middleware.apiAuthenticationRequired, function (req, res) {
        var token = accessTokens.createJwtToken(req, {}, 60*60*24*7);
        res.json(token);
    });

    function redirectWithFlash(req, res, url, err, flashObj) {

        if (err && Array.isArray(err)) req.flash('errors', err);
        if (err && !Array.isArray(err)) req.flash('error', err);

        if (typeof (flashObj) == 'object') {
            Object.keys(flashObj).forEach(function (k) {
                req.flash(k, flashObj[k]);
            });
        }
        res.redirect(url);

    }

    function renderInputs(req, res, view, locals) {
        locals = locals || {};
        locals.csrfToken = req.csrfToken && req.csrfToken();

        var flashs = (req.flash && req.flash()) || [];
        Object.keys(flashs).forEach(function (key) {
            locals[key] = flashs[key];
        });
        if (!locals.form)
            locals.form = locals.body;
        if (req.user)
            locals.user = req.user;

        locals.routes = routes;
        var filename = resolveViewPath(req,view, locals);
        res.render(filename, {
            inputs: locals
        });
    }

    function resolveViewPath(req,view, locals) {
        if (typeof (options.resolveViewPath) == 'function')
            view = options.resolveViewPath(req,view, locals);
        if (path.isAbsolute(view))
            return view;
        return path.join(viewsLocation, view);
    }

}